class BreakpointManager:
	def __init__(self, h_rule_functions):
		self.hammer_retq_breakpoints = []
		self.h_rule_breakpoints = []
		self.H_RULE_FUNCTIONS = h_rule_functions
		self.parse_virtual_functions = ["parse_action", "parse_choice", "parse_sequence", "parse_difference",
		"parse_many", "parse_and", "parse_attr_bool", "parse_bind", "parse_bits", "parse_butnot", "parse_charset",
		"parse_ch", "parse_end", "parse_endianness", "parse_epsilon", "parse_ignore", "parse_ignoreseq",
		"parse_indirect", "parse_int_range", "parse_not", "parse_nothing", "parse_optional", "parse_permutation",
		"parse_skip", "parse_seek", "parse_tell", "parse_token", "parse_unimplemented", "parse_put", "parse_get",
		"parse_whitespace", "parse_xor"] # TODO: use these in set_parse_virtual_breakpoints()


		self.h_do_parse = None
		self.h_packrat_parse = None
		self.perform_lowlevel_parse = None
		self.h_arena_malloc_raw = None

		self.init_parser_retq = None

		self.parse_action = None
		self.parse_choice = None
		self.parse_sequence = None
		self.parse_difference = None
		self.parse_many = None
		self.parse_and = None
		self.parse_attr_bool = None
		self.parse_bind = None
		self.parse_bits = None
		self.parse_butnot = None
		self.parse_charset = None
		self.parse_ch = None
		self.parse_end = None
		self.parse_endianness = None
		self.parse_epsilon = None
		self.parse_ignore = None
		self.parse_ignoreseq = None
		self.parse_indirect = None
		self.parse_int_range = None
		self.parse_not = None
		self.parse_nothing = None
		self.parse_optional = None
		self.parse_permutation = None
		self.parse_skip = None
		self.parse_seek = None
		self.parse_tell = None
		self.parse_token = None
		self.parse_unimplemented = None
		self.parse_put = None
		self.parse_get = None
		self.parse_whitespace = None
		self.parse_xor = None

		self.parse_virtual_ret_breakpoints = []
		#self.parse_action_ret = None
		#self.parse_choice_ret = None
		#self.parse_sequence_ret = None
		#self.parse_difference_ret = None
		#self.parse_many_ret = None
		#self.parse_and_ret = None
		#self.parse_attr_bool_ret = None
		#self.parse_bind_ret = None
		#self.parse_bits_ret = None
		#self.parse_butnot_ret = None
		#self.parse_charset_ret = None
		#self.parse_ch_ret = None
		#self.parse_end_ret = None
		#self.parse_endianness_ret = None
		#self.parse_epsilon_ret = None
		#self.parse_ignore_ret = None
		#self.parse_ignoreseq_ret = None
		#self.parse_indirect_ret = None
		#self.parse_int_range_ret = None
		#self.parse_not_ret = None
		#self.parse_nothing_ret = None
		#self.parse_optional_ret = None
		#self.parse_permutation_ret = None
		#self.parse_skip_ret = None
		#self.parse_seek_ret = None
		#self.parse_tell_ret = None
		#self.parse_token_ret = None
		#self.parse_unimplemented_ret = None
		#self.parse_put_ret = None
		#self.parse_get_ret = None
		#self.parse_whitespace_ret = None
		#self.parse_xor_ret = None

	def set_h_rule_breakpoints(self):
		for func in self.H_RULE_FUNCTIONS:
			func_retq = self.locate_retq(func)
			print("::: Setting breakpoint in " + func)
			self.h_rule_breakpoints.append(HRuleBreakpoint("*" + hex(func_retq)))

	def set_hammer_retq_breakpoints(self):
		plp_retq = self.locate_retq("perform_lowlevel_parse")
		print("::: Setting breakpoint in perform_lowlevel_parse")
		perform_lowlevel_parse_ret = PerformLowLevelParseRetBreakpoint("*" + hex(plp_retq))
		self.hammer_retq_breakpoints.append(perform_lowlevel_parse_ret)
		hpp_retq = self.locate_retq("h_packrat_parse")
		print("::: Setting breakpoint in h_packrat_parse")
		h_packrat_parse_ret = HPackratParseRetBreakpoint("*" + hex(hpp_retq))
		self.hammer_retq_breakpoints.append(h_packrat_parse_ret)
		hdp_retq = self.locate_retq("h_do_parse")
		print("::: Setting breakpoint in h_do_parse")
		# A potential problem in the future
		# h_do_parse() has multiple return statements, but compiled down to
		# a function with a single return instruction on the test machine
		# If the compiler decides to use multiple RETs, having only one
		# breakpoint set here will break the AST printing code.
		# Return values of h_do_parse might get missed.
		h_do_parse_ret = HDoParseRetBreakpoint("*" + hex(hdp_retq))
		self.hammer_retq_breakpoints.append(h_do_parse_ret)
		return self.hammer_retq_breakpoints

	def del_hammer_retq_breakpoints(self):
		for bp in self.hammer_retq_breakpoints:
			bp.delete()

	# Helper functions for finding the return instructions of particular functions of interest
	#TODO: rename "symbol" to "fn_name"
	def locate_retqs(self, symbol):
		arch = gdb.selected_frame().architecture()
		sym = gdb.lookup_symbol(symbol)[0]
		sym_address = int(sym.value().address)
		sym_block = sym.symtab.objfile.progspace.block_for_pc(sym_address)
		instructions = arch.disassemble(sym_address, sym_block.end)
		results = [ ins["addr"] for ins in instructions if ins["asm"].startswith("ret") ]
		return results

	def locate_retq(self, symbol):
		results = self.locate_retqs(symbol)
		return results[0]

	def set_parser_virtual_breakpoints(self):
		print("::: parse_action")
		self.parse_action = ParserVirtualBreakpoint("parse_action")
		print("::: parse_choice")
		self.parse_choice = ParserVirtualBreakpoint("parse_choice")
		print("::: parse_sequence")
		self.parse_sequence = ParserVirtualBreakpoint("parse_sequence")
		print("::: parse_difference")
		self.parse_difference = ParserVirtualBreakpoint("parse_difference")
		print("::: parse_many")
		self.parse_many = ParserVirtualBreakpoint("parse_many")
		print("::: parse_and")
		self.parse_and = ParserVirtualBreakpoint("parse_and")
		print("::: parse_attr_bool")
		self.parse_attr_bool = ParserVirtualBreakpoint("parse_attr_bool")
		print("::: parse_bind")
		self.parse_bind = ParserVirtualBreakpoint("parse_bind")
		print("::: parse_bits")
		self.parse_bits = ParserVirtualBreakpoint("parse_bits")
		print("::: parse_butnot")
		self.parse_butnot = ParserVirtualBreakpoint("parse_butnot")
		print("::: parse_charset")
		self.parse_charset = ParserVirtualBreakpoint("parse_charset")
		print("::: parse_ch")
		self.parse_ch = ParserVirtualBreakpoint("parse_ch")
		print("::: parse_end")
		self.parse_end = ParserVirtualBreakpoint("parse_end")
		print("::: parse_endianness")
		self.parse_endianness = ParserVirtualBreakpoint("parse_endianness")
		print("::: parse_epsilon")
		self.parse_epsilon = ParserVirtualBreakpoint("parse_epsilon")
		print("::: parse_ignore")
		self.parse_ignore = ParserVirtualBreakpoint("parse_ignore")
		print("::: parse_ignoreseq")
		self.parse_ignoreseq = ParserVirtualBreakpoint("parse_ignoreseq")
		print("::: parse_indirect")
		self.parse_indirect = ParserVirtualBreakpoint("parse_indirect")
		print("::: parse_int_range")
		self.parse_int_range = ParserVirtualBreakpoint("parse_int_range")
		print("::: parse_not")
		self.parse_not = ParserVirtualBreakpoint("parse_not")
		print("::: parse_nothing")
		self.parse_nothing = ParserVirtualBreakpoint("parse_nothing")
		print("::: parse_optional")
		self.parse_optional = ParserVirtualBreakpoint("parse_optional")
		print("::: parse_permutation")
		self.parse_permutation = ParserVirtualBreakpoint("parse_permutation")
		print("::: parse_skip")
		self.parse_skip = ParserVirtualBreakpoint("parse_skip")
		print("::: parse_seek")
		self.parse_seek = ParserVirtualBreakpoint("parse_seek")
		print("::: parse_tell")
		self.parse_tell = ParserVirtualBreakpoint("parse_tell")
		print("::: parse_token")
		self.parse_token = ParserVirtualBreakpoint("parse_token")
		print("::: parse_unimplemented")
		self.parse_unimplemented = ParserVirtualBreakpoint("parse_unimplemented")
		print("::: parse_put")
		self.parse_put = ParserVirtualBreakpoint("parse_put")
		print("::: parse_get")
		self.parse_get = ParserVirtualBreakpoint("parse_get")
		print("::: parse_whitespace")
		self.parse_whitespace = ParserVirtualBreakpoint("parse_whitespace")
		print("::: parse_xor")
		self.parse_xor = ParserVirtualBreakpoint("parse_xor")

	def del_parser_virtual_breakpoints(self):
		self.parse_action.delete()
		self.parse_choice.delete()
		self.parse_sequence.delete()
		self.parse_difference.delete()
		self.parse_many.delete()
		self.parse_and.delete()
		self.parse_attr_bool.delete()
		self.parse_bind.delete()
		self.parse_bits.delete()
		self.parse_butnot.delete()
		self.parse_charset.delete()
		self.parse_ch.delete()
		self.parse_end.delete()
		self.parse_endianness.delete()
		self.parse_epsilon.delete()
		self.parse_ignore.delete()
		self.parse_ignoreseq.delete()
		self.parse_indirect.delete()
		self.parse_int_range.delete()
		self.parse_not.delete()
		self.parse_nothing.delete()
		self.parse_optional.delete()
		self.parse_permutation.delete()
		self.parse_skip.delete()
		self.parse_seek.delete()
		self.parse_tell.delete()
		self.parse_token.delete()
		self.parse_unimplemented.delete()
		self.parse_put.delete()
		self.parse_get.delete()
		self.parse_whitespace.delete()
		self.parse_xor.delete()

	def set_parser_virtual_ret_breakpoints(self):
		for func_name in parse_virtual_functions:
			print(":::", func_name, "return breakpoint")
			func_retq = self.locate_retq(func_name)
			self.parse_virtual_ret_breakpoints.append(ParseVirtualRetBreakpoint("*"+hex(func_retq)))

	def del_parser_virtual_ret_breakpoints(self):
		for bp in parse_virtual_ret_breakpoints:
			bp.delete()

	#TODO: this probably should be merged into the H_RULE functions array instead
	def set_init_parser_breakpoint(self):
		i_p_retq = self.locate_retq("init_parser")
		self.init_parser_retq = InitParserBreakpoint("*"+hex(i_p_retq))

	def del_init_parser_breakpoint(self):
		self.init_parser_retq.delete()

	def set_hammer_breakpoints(self):
		print("::: h_do_parse")
		self.h_do_parse = HDoParseBreakpoint("h_do_parse")
		print("::: h_packrat_parse")
		self.h_packrat_parse = HPackratParseBreakpoint("h_packrat_parse")
		print("::: perform_lowlevel_parse")
		self.perform_lowlevel_parse = PerformLowLevelParseBreakpoint("perform_lowlevel_parse")
		print("::: h_arena_malloc_raw")
		self.h_arena_malloc_raw = HArenaMallocRawBreakpoint("h_arena_malloc_raw")

	def del_hammer_breakpoints(self):
		self.h_do_parse.delete()
		self.h_packrat_parse.delete()
		self.perform_lowlevel_parse.delete()
		self.h_arena_malloc_raw.delete()
