# These tests use parsers defined in the PDF parser, for expediency
# There's probably a way to test them in a general manner, but they'd need to be mocked out in process memory
import unittest

hparser_p_t = gdb.lookup_type("HParser").pointer()


class ChoiceEnvCreation(unittest.TestCase):
	# setup is also  done by parser-name-instrumentation.py and parser-type-instrumentation.py
	def setUp(self):
		self.parser_obj = top_level_parse.parsers_by_name("eol")[0]
		self.parser_env = top_level_parse.parser_decombinator.decompose_parser(self.parser_obj,top_level_parse)

	def test_parser_decombine(self):
		self.assertIsInstance(self.parser_env, ChoiceEnv)

	def test_member_parsers(self):
		self.assertIsNotNone(self.parser_env.member_parsers)
		for p in self.parser_env.member_parsers:
			self.assertIsInstance(p, Parser)

	def test_member_parsers_name(self):
		for p in self.parser_env.member_parsers:
			self.assertIsNotNone(p.name)

	def test_member_parsers_size(self):
		self.assertEqual(len(self.parser_env.member_parsers),3)


# Maybe separate tests for h_left, h_middle, etc.?
class IgnoreSeqEnvCreation(unittest.TestCase):
	# setup is also  done by parser-name-instrumentation.py and parser-type-instrumentation.py
	def setUp(self):
		self.parser_obj = top_level_parse.parsers_by_name("name")[0]
		self.parser_env = top_level_parse.parser_decombinator.decompose_parser(self.parser_obj,top_level_parse)

	def test_parser_decombine(self):
		self.assertIsInstance(self.parser_env, IgnoreSeqEnv)

	def test_member_parsers(self):
		self.assertIsNotNone(self.parser_env.member_parsers)
		for p in self.parser_env.member_parsers:
			self.assertIsInstance(p, Parser)

	def test_member_parsers_name(self):
		for p in self.parser_env.member_parsers:
			self.assertIsNotNone(p.name)

	def test_which(self):
		self.assertIsNotNone(self.parser_env.which)
		self.assertIsInstance(self.parser_env.which, int)

	def test_member_parsers_size(self):
		self.assertEqual(len(self.parser_env.member_parsers),2)


class ManyEnvCreation(unittest.TestCase):
	# setup is also  done by parser-name-instrumentation.py and parser-type-instrumentation.py
	def setUp(self):
		self.parser_obj = top_level_parse.parsers_by_name("digits")[0]
		self.parser_env = top_level_parse.parser_decombinator.decompose_parser(self.parser_obj,top_level_parse)

	def test_parser_decombine(self):
		self.assertIsInstance(self.parser_env, ManyEnv)

	def test_member_parser(self):
		self.assertIsNotNone(self.parser_env.p)
		self.assertIsInstance(self.parser_env.p, Parser)

	def test_member_parser_name(self):
		self.assertIsNotNone(self.parser_env.p.name)


# TODO: some parser combinators aren't represented in the H_RULE()s declared
#class NotEnvCreation(unittest.TestCase):
	#def setUp(self):
	#	self.parser_obj = top_level_parse.parsers_by_name("")[0]
	#	self.parser_env = top_level_parse.parser_decombinator.decompose_parser(self.parser_obj,top_level_parse)

	#def test_parser_decombine(self):
	#	self.assertIsInstance(self.parser_env, NotEnv)

	#def test_member_parser(self):
		# Member parser gets extracted from memory
	#	self.assertIsNotNone(self.parser_env.member_parser)
		# Parser gets named on creation
	#	self.assertIsNotNone(self.parser_env.member_parser.name)


class SequenceEnvCreation(unittest.TestCase):
	# setup is also  done by parser-name-instrumentation.py and parser-type-instrumentation.py
	def setUp(self):
		self.parser_obj = top_level_parse.parsers_by_name("comment")[0]
		self.parser_env = top_level_parse.parser_decombinator.decompose_parser(self.parser_obj,top_level_parse)

	def test_parser_decombine(self):
		self.assertIsInstance(self.parser_env, SequenceEnv)

	def test_member_parsers(self):
		self.assertIsNotNone(self.parser_env.member_parsers)
		for p in self.parser_env.member_parsers:
			self.assertIsInstance(p, Parser)

	def test_member_parsers_name(self):
		for p in self.parser_env.member_parsers:
			self.assertIsNotNone(p.name)

	def test_member_parsers_size(self):
		self.assertEqual(len(self.parser_env.member_parsers),3)

class TokenEnvCreation(unittest.TestCase):
	def setUp(self):
		self.parser_obj = top_level_parse.parsers_by_name("null")[0]
		self.parser_env = top_level_parse.parser_decombinator.decompose_parser(self.parser_obj,top_level_parse)
		self.token_type = gdb.lookup_type("uint8_t").pointer()

	def test_parser_decombine(self):
		self.assertIsInstance(self.parser_env, TokenEnv)

	def test_token(self):
		self.assertIsNotNone(self.parser_env.token)
		self.assertIsInstance(self.parser_env.token, gdb.Value)
		self.assertEqual(self.parser_env.token.type, self.token_type) # For gdb.Type objects, == checks for equivalence, and the "is" operator won't work

	def test_str_len(self):
		self.assertIsNotNone(self.parser_env.str_len)
		self.assertIsInstance(self.parser_env.str_len, int)

